import os
import sys
import base64
import hashlib
import tempfile
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

if sys.platform == "darwin":
    w0 = ''
else:
    'base64 command parameter'
    w0 = ' -w0'

def sign(challenge_message):
    command = 'echo "%s" | openssl rsautl -sign -inkey ~/.ssh/id_rsa | base64'+w0+' && echo -n ":" && cat ~/.ssh/id_rsa.pub | base64'+w0
    return os.popen(
        command % (challenge_message,)
    ).read().rstrip()

def verify(challenge_response):
    try:
        sig, pub = challenge_response.split(':')

        pubkey = base64.b64decode(pub)
        key = serialization.load_ssh_public_key(pubkey, default_backend())
        pemkey = key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        rawkey = key.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH
        )
        b64pubkey = str(rawkey, 'utf-8').split(' ')[-1]
        sha256 = hashlib.sha256()
        sha256.update(base64.b64decode(b64pubkey))
        b64fingerprint = base64.b64encode(sha256.digest())

        keyprint = {'fingerprint': b64fingerprint, 'pem': pemkey }


        f = tempfile.NamedTemporaryFile()
        challenge_recovered = None
        try:
            f.write(pemkey)
            f.seek(0)
            command = 'echo "%s" | base64 -d'+w0+' | openssl rsautl -verify -inkey %s -pubin'
            challenge_recovered = os.popen(
                command % (sig, f.name)
            ).read().rstrip()
        finally:
            f.close()

        if challenge_recovered is not None:
            return {
                'success': True,
                'status': 'Message recovered.',
                'info': 'Check if it matches the challenge message.',
                'recovered_challenge_message': challenge_recovered,
                'key': keyprint,
            }
        else:
            return {
                'success': False,
                'status': 'Verification failed.',
                'info': 'Could not succeed with command "openssl rsautl -verify -inkey".',
                'recovered_challenge_message': challenge_recovered or 'UNKNOWN',
                'key': keyprint,
            }
    except Exception as e:
        return {
            'success': False,
            'status': 'Verification failed with error.',
            'error': repr(e),
            'recovered_challenge_message': 'UNKNOWN',
        }




#!/usr/bin/python3
import json, os, re, argparse, pathlib
from collections import defaultdict
import subprocess

from ident import verify

class SignFileExtensionError(Exception):
    """The signatures file must have .sign extension."""

class SignFilenameInvalidError(Exception):
    """The hash contained before .sign does not match sha256sum of .manifest."""

class ManifestNotFound(Exception):
    """Manifest (.manifest) file not found."""

class ManifestMismatchFolderError(Exception):
    """Manifest (.manifest) file hashes do not match location's content hashes."""

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dir', help='Source directory.')
    parser.add_argument('-s', '--sig', help='Signatures file.')
    parser.add_argument('-m', '--man', help='Manifest file.')
    args = parser.parse_args()

    CWD = pathlib.Path(os.getcwd())
    DIR = pathlib.Path(args.dir)
    SIG = pathlib.Path(args.sig)
    MAN = pathlib.Path(args.man or '.manifest')

    directory = os.path.join(CWD, DIR)
    manifest = os.path.join(CWD, MAN)
    signatures = os.path.join(CWD, SIG)

    # Reading signatures file content
    if not args.sig.endswith('.sign'):
        raise SignFileExtensionError

    challenge_message = args.sig.split('.sign', 1)[0]

    # Verify challenge message (must be equal to .manifest hash)
    if os.path.exists(manifest):
        manihash = subprocess.getoutput("sha256sum %s | awk '{print $1}'" % manifest)

        if challenge_message == manihash:
            print("[OK] Manifest (.manifest) hash is in .sign filename.")
        else:
            print("[--] Wrong manifest hash: mismating .sign filename.")
            raise SignFilenameInvalidError
    else:
        raise ManifestNotFound

    # Verify that manifest is correct, and reflects files:
    manifest_file_hashes = open(manifest).read().splitlines()
    prefix = f'location="{directory}"; '
    command = prefix + """find "$location" -type f -exec sha256sum {} \; | awk '{print $1}' | LC_ALL=C sort -d"""
    folder_file_hashes = subprocess.getoutput(command).splitlines()

    if manifest_file_hashes == folder_file_hashes:
        print(f"[OK] Manifest (.manifest) hashes match folder's file hashes.")
    else:
        print(f"[--] Manifest (.manifest) file hashes do not match location's ({directory}) content hashes.")
        raise ManifestMismatchFolderError

    # Verify Signatures
    print(f"\nVerifying signatures one by one:")
    successes_list = []

    for line in open(signatures, 'r').read().splitlines():
        name, sign = line.split(',', 1)
        result = verify(sign)

        line_success = False
        if result['success']:
            recovered_message = result.get('recovered_challenge_message')
            if recovered_message:
                if recovered_message == challenge_message:
                    line_success = True

        if line_success:
            print(f'- {name}: [YES]')
            successes_list += [True]
        else:
            print(f'- {name}: [NO]')
            successes_list += [False]

        if all(successes_list):
            print("All signatures are valid.")
            print(f"\nFinal hash is correct, and is as folder's name:\n{DIR}")
            print("(It is safe to write it to blockchain, as proof of data witness.)")
        elif any(successes_list):
            print("Not all signatures are valid.")
        else:
            print("No valid signatures found.")

if __name__ == '__main__':
    main()
